#! /bin/bash
#########################################################################
#									#
#	lxcu-copy.sh is automatically generated,			#
#		please do not modify!					#
#									#
#########################################################################

#########################################################################
#									#
# Script ID: lxcu-copy.sh						#
# Author: Copyright (C) 2020-2023  Mark Grant				#
#									#
# Released under the GPLv3 only.					#
# SPDX-License-Identifier: GPL-3.0-only					#
#									#
# Purpose:								#
# A script to copy an existing container to a new container. The only	#
# extra thing this does compared to an lxc-copy is that it will create	#
# any extra lxcu bridge config files.					#
#									#
# Syntax:	lxcu-copy.sh [ -h or --help				#
#			|| -n or --from-name 'container name' copy from	#
#			|| -N or --to-name 'container name' copy to	#
#			|| -V or --version ]				#
#									#
# Exit codes used:-							#
# Bash standard Exit Codes:	0 - success				#
#				1 - general failure			#
# User-defined exit code range is 64 - 113				#
#	C/C++ Semi-standard exit codes from sysexits.h range is 64 - 78	#
#		EX_USAGE	64	command line usage error	#
#		EX_DATAERR	65	data format error		#
#		EX_NOINPUT	66	cannot open input		#
#		EX_NOUSER	67	addressee unknown		#
#		EX_NOHOST	68	host name unknown		#
#		EX_UNAVAILABLE	69	service unavailable		#
#		EX_SOFTWARE	70	internal software error		#
#		EX_OSERR	71	system error (e.g., can't fork)	#
#		EX_OSFILE	72	critical OS file missing	#
#		EX_CANTCREAT	73	can't create (user) output file	#
#		EX_IOERR	74	input/output error		#
#		EX_TEMPFAIL	75	temp failure; user is invited	#
#					to retry			#
#		EX_PROTOCOL	76	remote error in protocol	#
#		EX_NOPERM	77	permission denied		#
#		EX_CONFIG	78	configuration error		#
#	User-defined (here) exit codes range 79 - 113:			#
#		None							#
#									#
# Further Info:								#
#									#
#########################################################################

#########################################################################
#									#
# Changelog								#
#									#
# Date		Author	Version	Description				#
#									#
# 21/11/2020	MG	1.0.1	Created.				#
# 13/02/2021	MG	1.0.2	Prepend this package executable		#
#				invocations with bindir.		#
# 03/09/2021	MG	1.0.3	With lxc v4 lxc start and attach need	#
#				slightly different privileges. The	#
#				altered commands also work pre-v4.	#
# 28/11/2021	MG	1.0.4	Tighten SPDX tag.			#
# 08/01/2022	MG	1.0.5	Use lxcu-attach.sh AOT systemd-run	#
#				lxc-attach				#
# 28/04/2023	MG	1.0.6	Remove unused variable outputprefix.	#
#				Double quote variables to prevent word	#
#				splitting and globbing.			#
# 13/05/2023	MG	1.0.7	Improve config file read error handling.#
#				Centralise error exit in		#
#				std_cmd_err_handler().			#
#				Cleanup on error exit.			#
#				Validate inputs.			#
#									#
#########################################################################


set -o pipefail


##################
# Init variables #
##################
readonly version=1.0.7				# Script version
readonly packageversion=1.1.6		# Package version

readonly etclocation=/etc/lxcu		# Path to config directory

# Ensure environment variables are set.
export XDG_RUNTIME_DIR="/run/user/$UID"
export DBUS_SESSION_BUS_ADDRESS="unix:path=${XDG_RUNTIME_DIR}/bus"


#############
# Functions #
#############

# -h --help output.
# No parameters
# No return value
usage()
{
cat << EOF
Usage is $0 [options]
	-h or --help Display this help information
	-n or --from-name 'container name'
	-N or --to-name 'container name'
	-V or --version Print script version information
EOF
}

# Standard function to emit messages depending on various parameters.
# Parameters -	$1 What:-	The message to emit.
#		$2 Where:-	stdout == 0
#				stderr == 1
# No return value.
output()
{
	if (( !$2 )); then
		printf "%s\n" "$1"
	else
		printf "%s\n" "$1" 1>&2
	fi
}

# Standard function to tidy up and return exit code.
# Parameters - 	$1 is the exit code.
# No return value.
script_exit()
{
	exit "$1"
}

# Standard function to test command error and exit if non-zero.
# Parameters - 	$1 is the exit code, (normally $? from the preceeding command).
# No return value.
std_cmd_err_handler()
{
	if (( $1 )); then
		if [[ -n "$toname" ]]; then
			if (( $(lxc-ls -1 --running --filter="^$toname$" \
				| wc -l) )); then
				lxc-stop -n "$toname"
			fi
			if (( $(lxc-ls -1 --filter="^$toname$" | wc -l) )); then
				lxc-destroy -n "$toname"
			fi
		fi
		script_exit "$1"
	fi
}

# Standard trap exit function.
# No parameters.
# No return value.
trap_exit()
{
	local -i exit_code=$?
	local msg

	msg="Script terminating with exit code $exit_code due to trap received."
	output "$msg" 1
	std_cmd_err_handler $exit_code
}

# Setup trap
trap trap_exit SIGHUP SIGINT SIGQUIT SIGTERM

# Process the config file just for the parameters of interest.
# Parameters - None
# No return value.
proc_config_file()
{
	local input=()
	local msg
	local oldIFS=$IFS

	IFS="="

	if [[ ! -e "$etclocation/lxcu.conf" || ! -f "$etclocation/lxcu.conf" \
		|| ! -r "$etclocation/lxcu.conf" ]]; then
		msg="$etclocation/lxcu.conf must exist, be a file and be"
		msg+=" readable."
		output "$msg" 1
		std_cmd_err_handler 66
	fi

	exec 3<"$etclocation/lxcu.conf"
	std_cmd_err_handler $?
	while read -u3 -ra input; do
		case ${input[0]} in
		bridge)
			bridge=${input[1]}
			;;
		contlocation)
			contlocation=$(echo "${input[1]}" | envsubst)
			;;
		esac
	done
	exec 3<&-

	IFS=$oldIFS
}

# Process command line arguments with GNU getopt.
# Parameters -	$1 is the command line.
# No return value.
proc_CL()
{
	local GETOPTTEMP
	local msg
	local tmp

	tmp="getopt -o hn:N:V --long dist:,help,from-name:,to-name:,version"
	GETOPTTEMP=$($tmp -n "$0" -- "$@")
	std_cmd_err_handler $?

	eval set -- "$GETOPTTEMP"
	std_cmd_err_handler $?

	while true; do
		case "$1" in
		-h|--help)
			usage
			shift
			script_exit 0
			;;
		-n|--from-name)
			fromname=$2
			shift 2
			;;
		-N|--to-name)
			# Target container must not exist.
			if (( $(lxc-ls -1 --filter="^$2$" | wc -l) )); then
				msg="Target container $2 must not exist."
				output "$msg" 1
				std_cmd_err_handler 64
			fi
			toname=$2
			shift 2
			;;
		-V|--version)
			printf "%s Script version %s\n" "$0" $version
			printf "%s Package version %s\n" "$0" $packageversion
			shift
			script_exit 0
			;;
		--)	shift
			break
			;;
		*)	output "Internal error." 1
			std_cmd_err_handler 64
			;;
		esac
	done

	# Script does not accept other arguments.
	if (( $# > 0 )); then
		output "Invalid argument." 1
		std_cmd_err_handler 64
	fi

	# Both from name and to name must be specified.
	if [[ -z $fromname || -z $toname ]]; then
		msg="Both from name and to name must be specified."
		output "$msg" 1
		std_cmd_err_handler 64
	fi

	# Source container must exist but must not be running.
	if (( ! $(lxc-ls -1 --filter="^$fromname$" | wc -l) )); then
		output "Source container $fromname must exist." 1
		std_cmd_err_handler 64
	fi
	if (( $(lxc-ls -1 --running --filter="^$fromname$" | wc -l) )); then
		output "Source container $fromname must not be running." 1
		std_cmd_err_handler 64
	fi
}


########
# Main #
########

proc_config_file

proc_CL "$@"

lxc-copy -n "$fromname" -N "$toname"
std_cmd_err_handler $?

# Any bridge to provide a config file for?
if [[ -n $bridge ]]; then
	cp -v "$contlocation/$fromname/config-$bridge" \
		"$contlocation/$toname"
	std_cmd_err_handler $?
	sed -i -e "/lxc.rootfs.path = dir:/s/$fromname/$toname/" \
		"$contlocation/$toname/config-$bridge"
	std_cmd_err_handler $?
	sed -i -e "/lxc.uts.name = /s/$fromname/$toname/" \
		"$contlocation/$toname/config-$bridge"
	std_cmd_err_handler $?
fi

# Correct host name on target container.
/usr/bin/lxcu-start.sh -n "$toname"
std_cmd_err_handler $?

/usr/bin/lxcu-attach.sh -n "$toname" \
	-- sudo sed -- -i -e "s/$fromname/$toname/" /etc/hostname
std_cmd_err_handler $?

/usr/bin/lxcu-attach.sh -n "$toname" \
	-- sudo sed -- -i -e "s/$fromname/$toname/" /etc/hosts
std_cmd_err_handler $?

lxc-stop -n "$toname"
std_cmd_err_handler $?

# And exit.
script_exit 0

