#! /bin/bash
#########################################################################
#									#
#	lxcu-start.sh is automatically generated,			#
#		please do not modify!					#
#									#
#########################################################################

#########################################################################
#									#
# Script ID: lxcu-start.sh						#
# Author: Copyright (C) 2020-2023  Mark Grant				#
#									#
# Released under the GPLv3 only.					#
# SPDX-License-Identifier: GPL-3.0-only					#
#									#
# Purpose:								#
# A script to start a linux container.					#
#									#
# Syntax:	lxcu-start.sh [ -h or --help				#
#			|| -n or --name 'container name'		#
#			|| -V or --version ]				#
#									#
# Exit codes used:-							#
# Bash standard Exit Codes:	0 - success				#
#				1 - general failure			#
# User-defined exit code range is 64 - 113				#
#	C/C++ Semi-standard exit codes from sysexits.h range is 64 - 78	#
#		EX_USAGE	64	command line usage error	#
#		EX_DATAERR	65	data format error		#
#		EX_NOINPUT	66	cannot open input		#
#		EX_NOUSER	67	addressee unknown		#
#		EX_NOHOST	68	host name unknown		#
#		EX_UNAVAILABLE	69	service unavailable		#
#		EX_SOFTWARE	70	internal software error		#
#		EX_OSERR	71	system error (e.g., can't fork)	#
#		EX_OSFILE	72	critical OS file missing	#
#		EX_CANTCREAT	73	can't create (user) output file	#
#		EX_IOERR	74	input/output error		#
#		EX_TEMPFAIL	75	temp failure; user is invited	#
#					to retry			#
#		EX_PROTOCOL	76	remote error in protocol	#
#		EX_NOPERM	77	permission denied		#
#		EX_CONFIG	78	configuration error		#
#	User-defined (here) exit codes range 79 - 113:			#
#		None							#
#									#
# Further Info:								#
#									#
#########################################################################

#########################################################################
#									#
# Changelog								#
#									#
# Date		Author	Version	Description				#
#									#
# 25/09/2020	MG	1.0.1	Created.				#
# 05/10/2020	MG	1.0.2	Add wait for network initialisation.	#
# 14/11/2020	MG	1.0.3	Wait for server-dependency service to	#
#				complete.				#
# 03/09/2021	MG	1.0.4	With lxc v4 lxc start and attach need	#
#				slightly different privileges. The	#
#				altered commands also work pre-v4.	#
# 28/11/2021	MG	1.0.5	Tighten SPDX tag.			#
# 08/01/2022	MG	1.0.6	Use lxcu-attach.sh AOT systemd-run	#
#				lxc-attach				#
# 05/05/2023	MG	1.0.7	Remove unused variable outputprefix.	#
#				Double quote variables to prevent word	#
#				splitting and globbing.			#
#				Fix ethernet_bridge typo.		#
#				Remove unused variable vni.		#
#				Use grep -c AOT grep | wc -l		#
#				Remove unnecessary $ from variable	#
#				names in arithmetic expressions.	#
# 16/05/2023	MG	1.0.8	Improve config file error handling.	#
#				Container must exist but must not be	#
#				running.				#
#									#
#########################################################################


set -o pipefail


##################
# Init variables #
##################
readonly version=1.0.8				# Script version
readonly packageversion=1.1.6		# Package version

readonly etclocation=/etc/lxcu		# Path to config directory

# Ensure environment variables are set.
export XDG_RUNTIME_DIR="/run/user/$UID"
export DBUS_SESSION_BUS_ADDRESS="unix:path=${XDG_RUNTIME_DIR}/bus"


#############
# Functions #
#############

# -h --help output.
# No parameters
# No return value
usage()
{
cat << EOF
Usage is $0 [options]
	-h or --help Display this help information
	-n or --name 'container name'
	-V or --version Print script version information
EOF
}

# Standard function to emit messages depending on various parameters.
# Parameters -	$1 What:-	The message to emit.
#		$2 Where:-	stdout == 0
#				stderr == 1
# No return value.
output()
{
	if (( !$2 )); then
		printf "%s\n" "$1"
	else
		printf "%s\n" "$1" 1>&2
	fi
}

# Standard function to tidy up and return exit code.
# Parameters - 	$1 is the exit code.
# No return value.
script_exit()
{
	exit "$1"
}

# Standard function to test command error and exit if non-zero.
# Parameters - 	$1 is the exit code, (normally $? from the preceeding command).
# No return value.
std_cmd_err_handler()
{
	if (( $1 )); then
		script_exit "$1"
	fi
}

# Standard trap exit function.
# No parameters.
# No return value.
trap_exit()
{
	local -i exit_code=$?
	local msg

	msg="Script terminating with exit code $exit_code due to trap received."
	output "$msg" 1
	std_cmd_err_handler $exit_code
}

# Setup trap
trap trap_exit SIGHUP SIGINT SIGQUIT SIGTERM

# Process the config file just for the parameters of interest.
# Parameters - None
# No return value.
proc_config_file()
{
	local input=()
	local msg
	local oldIFS=$IFS

	IFS="="

	if [[ ! -e "$etclocation/lxcu.conf" || ! -f "$etclocation/lxcu.conf" \
		|| ! -r "$etclocation/lxcu.conf" ]]; then
		msg="$etclocation/lxcu.conf must exist, be a file and be"
		msg+=" readable."
		output "$msg" 1
		std_cmd_err_handler 66
	fi

	exec 3<"$etclocation/lxcu.conf"
	std_cmd_err_handler $?
	while read -u3 -ra input; do
		case ${input[0]} in
		contlocation)
			contlocation=$(echo "${input[1]}" | envsubst)
			;;
		bridge)
			bridge=${input[1]}
			;;
		ethernetprefix)
			ethernet_prefix=${input[1]}
			;;
		esac
	done
	exec 3<&-

	IFS=$oldIFS
}

# Process command line arguments with GNU getopt.
# Parameters -	$1 is the command line.
# No return value.
proc_CL()
{
	local GETOPTTEMP
	local msg
	local tmp

	tmp="getopt -o hn:V --long help,name:,version"
	GETOPTTEMP=$($tmp -n "$0" -- "$@")
	std_cmd_err_handler $?

	eval set -- "$GETOPTTEMP"
	std_cmd_err_handler $?

	while true; do
		case "$1" in
		-h|--help)
			usage
			shift
			script_exit 0
			;;
		-n|--name)
			name=$2
			shift 2
			;;
		-V|--version)
			printf "%s Script version %s\n" "$0" $version
			printf "%s Package version %s\n" "$0" $packageversion
			shift
			script_exit 0
			;;
		--)	shift
			break
			;;
		*)	output "Internal error." 1
			std_cmd_err_handler 64
			;;
		esac
	done

	# Script does not accept other arguments.
	if (( $# > 0 )); then
		output "Invalid argument." 1
		std_cmd_err_handler 64
	fi

	# Container name must be specified.
	if [[ -z $name ]]; then
		output "Container name must be specified." 1
		std_cmd_err_handler 64
	fi
	# Container must exist but must not be running.
	if (( ! $(lxc-ls -1 --filter="^$name$" | wc -l) )); then
		output "Container $name must exist." 1
		std_cmd_err_handler 64
	fi
	if (( $(lxc-ls -1 --running --filter="^$name$" | wc -l) )); then
		output "Container $name must not be running." 1
		std_cmd_err_handler 64
	fi
}


########
# Main #
########

proc_config_file

proc_CL "$@"

cmd="systemd-run --quiet --user --scope -p \"Delegate=yes\" lxc-start -n $name"

if [[ -n $bridge ]]; then
	# Enable / disable bridge VNI depending on hard-wired ethernet port
	# status.
	if (($(ip link show | grep -c "$ethernet_prefix.*$bridge.*state UP")))
	then
		cmd+=" -f $contlocation/$name/config-$bridge"
	fi
fi

eval "$cmd"
std_cmd_err_handler $?

lxc-wait -n "$name" -s RUNNING
std_cmd_err_handler $?

# Allow networking to initialise, otherwise problems encountered with
# Temporary failure resolving ....
declare -i count
declare -i max

((max=30))

until (/usr/bin/lxcu-attach.sh -n "$name" \
	-- ping -- -c 2 www.google.com &>/dev/null); do
	err=$?
	(( count++ ))
	if (( count == max )); then
		output "Networking not functioning." 1
		std_cmd_err_handler $err
	fi
	sleep 2
done

# Wait for the server dependency service to complete.
until ! (/usr/bin/lxcu-attach.sh -n "$name" \
	-- sudo systemctl is-active -- -q server-dependency); do
	sleep 2
done

# And exit.
script_exit 0

